/// Interegular transition key.
pub type TransitionKey = u32;

/// Token content.
pub type Token = String;

/// Token identifier.
pub type TokenId = u32;

/// Interegular state.
pub type State = u32;
